/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "hashfactory.h"
#include "doublemd5.h"
#include "doublemd5help.h"
#include "hex.h"

uint32_t DoubleMD5::inited = DoubleMD5::initCalc();

#ifdef ARC_x86

void doubleMd5Func2(__m128i *pw, __m128i *hash)
{
	doubleMd5Func(pw, hash);
}

void DoubleMD5::calcSSE2(vector4 *pw, vector4 *hash)
{
	doubleMd5Func2((__m128i*) pw, (__m128i*) hash + __DOUBLE_MD5_INTERLACED__);
	if (__DOUBLE_MD5_INTERLACED__ > 0) { ((__m128i*) hash)[0 * __DOUBLE_MD5_INTERLACED__ + 0] = _mm_set1_epi32(32); }
	if (__DOUBLE_MD5_INTERLACED__ > 1) { ((__m128i*) hash)[0 * __DOUBLE_MD5_INTERLACED__ + 1] = _mm_set1_epi32(32); }
	if (__DOUBLE_MD5_INTERLACED__ > 2) { ((__m128i*) hash)[0 * __DOUBLE_MD5_INTERLACED__ + 2] = _mm_set1_epi32(32); }
	if (__DOUBLE_MD5_INTERLACED__ > 3) { ((__m128i*) hash)[0 * __DOUBLE_MD5_INTERLACED__ + 3] = _mm_set1_epi32(32); }
	if (__DOUBLE_MD5_INTERLACED__ > 0) { ((__m128i*) hash)[9 * __DOUBLE_MD5_INTERLACED__ + 0] = _mm_set1_epi32(0x80); }
	if (__DOUBLE_MD5_INTERLACED__ > 1) { ((__m128i*) hash)[9 * __DOUBLE_MD5_INTERLACED__ + 1] = _mm_set1_epi32(0x80); }
	if (__DOUBLE_MD5_INTERLACED__ > 2) { ((__m128i*) hash)[9 * __DOUBLE_MD5_INTERLACED__ + 2] = _mm_set1_epi32(0x80); }
	if (__DOUBLE_MD5_INTERLACED__ > 3) { ((__m128i*) hash)[9 * __DOUBLE_MD5_INTERLACED__ + 3] = _mm_set1_epi32(0x80); }
	__m128i vh, vl, const_0x0f, const_0x09, const_0x27, const_0x30;

	const_0x0f = _mm_set1_epi32(0x0f0f0f0f);
	const_0x09 = _mm_set1_epi32(0x09090909);
	const_0x27 = _mm_set1_epi32(0x27272727);
	const_0x30 = _mm_set1_epi32(0x30303030);

	#define x(i,j)  ((__m128i*) hash)[(i) * __DOUBLE_MD5_INTERLACED__ + (j)]
	if (__DOUBLE_MD5_INTERLACED__ > 3) { CONVERT_TO_HEX(x(4,3), x(7,3), x(8,3), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 2) { CONVERT_TO_HEX(x(4,2), x(7,2), x(8,2), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 1) { CONVERT_TO_HEX(x(4,1), x(7,1), x(8,1), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 0) { CONVERT_TO_HEX(x(4,0), x(7,0), x(8,0), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 3) { CONVERT_TO_HEX(x(3,3), x(5,3), x(6,3), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 2) { CONVERT_TO_HEX(x(3,2), x(5,2), x(6,2), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 1) { CONVERT_TO_HEX(x(3,1), x(5,1), x(6,1), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 0) { CONVERT_TO_HEX(x(3,0), x(5,0), x(6,0), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 3) { CONVERT_TO_HEX(x(2,3), x(3,3), x(4,3), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 2) { CONVERT_TO_HEX(x(2,2), x(3,2), x(4,2), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 1) { CONVERT_TO_HEX(x(2,1), x(3,1), x(4,1), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 0) { CONVERT_TO_HEX(x(2,0), x(3,0), x(4,0), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 3) { CONVERT_TO_HEX(x(1,3), x(1,3), x(2,3), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 2) { CONVERT_TO_HEX(x(1,2), x(1,2), x(2,2), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 1) { CONVERT_TO_HEX(x(1,1), x(1,1), x(2,1), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	if (__DOUBLE_MD5_INTERLACED__ > 0) { CONVERT_TO_HEX(x(1,0), x(1,0), x(2,0), vl, vh, const_0x0f, const_0x09, const_0x27, const_0x30); }
	#undef x
	doubleMd5Func2((__m128i*) hash, (__m128i*) hash);
}

#endif

void doubleMd5Func_Norm(vector4 *pw, vector4 *hash)
{
	uint32_t a = 0x67452301, b = 0xefcdab89, c = 0x98badcfe, d = 0x10325476;

	#define MD5_FF(a, b, c, d, C, x, s) a += (d ^ (b & (c ^ d))) + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_GG(a, b, c, d, C, x, s) a += (c ^ (d & (b ^ c))) + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_HH(a, b, c, d, C, x, s) a += (b ^ c ^ d)         + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;
	#define MD5_II(a, b, c, d, C, x, s) a += (c ^ (b | ~d))      + C + FROM_LITTLE_ENDIAN_32(x); a = (a << s) | (a >> (32 - s)); a += b;

	#define x0   pw[1].data[0]
	#define x1   pw[2].data[0]
	#define x2   pw[3].data[0]
	#define x3   pw[4].data[0]
	#define x4   pw[5].data[0]
	#define x5   pw[6].data[0]
	#define x6   pw[7].data[0]
	#define x7   pw[8].data[0]
	#define x8   pw[9].data[0]
	#define x9   0
	#define x10  0
	#define x11  0
	#define x12  0
	#define x13  0
	#define x14  pw[0].data[0]
	#define x15  0

	x14 <<= 3;

	MD5_FF(a, b, c, d, 0xd76aa478,  x0,  7); MD5_FF(d, a, b, c, 0xe8c7b756,  x1, 12); MD5_FF(c, d, a, b, 0x242070db,  x2, 17); MD5_FF(b, c, d, a, 0xc1bdceee,  x3, 22);
	MD5_FF(a, b, c, d, 0xf57c0faf,  x4,  7); MD5_FF(d, a, b, c, 0x4787c62a,  x5, 12); MD5_FF(c, d, a, b, 0xa8304613,  x6, 17); MD5_FF(b, c, d, a, 0xfd469501,  x7, 22);
	MD5_FF(a, b, c, d, 0x698098d8,  x8,  7); MD5_FF(d, a, b, c, 0x8b44f7af,  x9, 12); MD5_FF(c, d, a, b, 0xffff5bb1, x10, 17); MD5_FF(b, c, d, a, 0x895cd7be, x11, 22);
	MD5_FF(a, b, c, d, 0x6b901122, x12,  7); MD5_FF(d, a, b, c, 0xfd987193, x13, 12); MD5_FF(c, d, a, b, 0xa679438e, x14, 17); MD5_FF(b, c, d, a, 0x49b40821, x15, 22);

	MD5_GG(a, b, c, d, 0xf61e2562,  x1,  5); MD5_GG(d, a, b, c, 0xc040b340,  x6,  9); MD5_GG(c, d, a, b, 0x265e5a51, x11, 14); MD5_GG(b, c, d, a, 0xe9b6c7aa,  x0, 20);
	MD5_GG(a, b, c, d, 0xd62f105d,  x5,  5); MD5_GG(d, a, b, c, 0x02441453, x10,  9); MD5_GG(c, d, a, b, 0xd8a1e681, x15, 14); MD5_GG(b, c, d, a, 0xe7d3fbc8,  x4, 20);
	MD5_GG(a, b, c, d, 0x21e1cde6,  x9,  5); MD5_GG(d, a, b, c, 0xc33707d6, x14,  9); MD5_GG(c, d, a, b, 0xf4d50d87,  x3, 14); MD5_GG(b, c, d, a, 0x455a14ed,  x8, 20);
	MD5_GG(a, b, c, d, 0xa9e3e905, x13,  5); MD5_GG(d, a, b, c, 0xfcefa3f8,  x2,  9); MD5_GG(c, d, a, b, 0x676f02d9,  x7, 14); MD5_GG(b, c, d, a, 0x8d2a4c8a, x12, 20);

	MD5_HH(a, b, c, d, 0xfffa3942,  x5,  4); MD5_HH(d, a, b, c, 0x8771f681,  x8, 11); MD5_HH(c, d, a, b, 0x6d9d6122, x11, 16); MD5_HH(b, c, d, a, 0xfde5380c, x14, 23);
	MD5_HH(a, b, c, d, 0xa4beea44,  x1,  4); MD5_HH(d, a, b, c, 0x4bdecfa9,  x4, 11); MD5_HH(c, d, a, b, 0xf6bb4b60,  x7, 16); MD5_HH(b, c, d, a, 0xbebfbc70, x10, 23);
	MD5_HH(a, b, c, d, 0x289b7ec6, x13,  4); MD5_HH(d, a, b, c, 0xeaa127fa,  x0, 11); MD5_HH(c, d, a, b, 0xd4ef3085,  x3, 16); MD5_HH(b, c, d, a, 0x04881d05,  x6, 23);
	MD5_HH(a, b, c, d, 0xd9d4d039,  x9,  4); MD5_HH(d, a, b, c, 0xe6db99e5, x12, 11); MD5_HH(c, d, a, b, 0x1fa27cf8, x15, 16); MD5_HH(b, c, d, a, 0xc4ac5665,  x2, 23);

	MD5_II(a, b, c, d, 0xf4292244,  x0,  6); MD5_II(d, a, b, c, 0x432aff97,  x7, 10); MD5_II(c, d, a, b, 0xab9423a7, x14, 15); MD5_II(b, c, d, a, 0xfc93a039,  x5, 21);
	MD5_II(a, b, c, d, 0x655b59c3, x12,  6); MD5_II(d, a, b, c, 0x8f0ccc92,  x3, 10); MD5_II(c, d, a, b, 0xffeff47d, x10, 15); MD5_II(b, c, d, a, 0x85845dd1,  x1, 21);
	MD5_II(a, b, c, d, 0x6fa87e4f,  x8,  6); MD5_II(d, a, b, c, 0xfe2ce6e0, x15, 10); MD5_II(c, d, a, b, 0xa3014314,  x6, 15); MD5_II(b, c, d, a, 0x4e0811a1, x13, 21);
	MD5_II(a, b, c, d, 0xf7537e82,  x4,  6); MD5_II(d, a, b, c, 0xbd3af235, x11, 10); MD5_II(c, d, a, b, 0x2ad7d2bb,  x2, 15); MD5_II(b, c, d, a, 0xeb86d391,  x9, 21);

	#undef MD5_FF
	#undef MD5_GG
	#undef MD5_HH
	#undef MD5_II

	#undef x0
	#undef x1
	#undef x2
	#undef x3
	#undef x4
	#undef x5
	#undef x6
	#undef x7
	#undef x8
	#undef x9
	#undef x10
	#undef x11
	#undef x12
	#undef x13
	#undef x14
	#undef x15

	hash[0].data[0] = TO_LITTLE_ENDIAN_32(a + 0x67452301);
	hash[1].data[0] = TO_LITTLE_ENDIAN_32(b + 0xefcdab89);
	hash[2].data[0] = TO_LITTLE_ENDIAN_32(c + 0x98badcfe);
	hash[3].data[0] = TO_LITTLE_ENDIAN_32(d + 0x10325476);
}

void DoubleMD5::calcNorm(vector4 *pw, vector4 *hash)
{
	doubleMd5Func_Norm(pw, hash + 1);

	hash[9].data[0]   = TO_LITTLE_ENDIAN_32(0x80);
	hash[8].data16[1] = CONST_BYTE_TO_HEX[hash[4].data8[3]];
	hash[8].data16[0] = CONST_BYTE_TO_HEX[hash[4].data8[2]];
	hash[7].data16[1] = CONST_BYTE_TO_HEX[hash[4].data8[1]];
	hash[7].data16[0] = CONST_BYTE_TO_HEX[hash[4].data8[0]];

	hash[6].data16[1] = CONST_BYTE_TO_HEX[hash[3].data8[3]];
	hash[6].data16[0] = CONST_BYTE_TO_HEX[hash[3].data8[2]];
	hash[5].data16[1] = CONST_BYTE_TO_HEX[hash[3].data8[1]];
	hash[5].data16[0] = CONST_BYTE_TO_HEX[hash[3].data8[0]];

	hash[4].data16[1] = CONST_BYTE_TO_HEX[hash[2].data8[3]];
	hash[4].data16[0] = CONST_BYTE_TO_HEX[hash[2].data8[2]];
	hash[3].data16[1] = CONST_BYTE_TO_HEX[hash[2].data8[1]];
	hash[3].data16[0] = CONST_BYTE_TO_HEX[hash[2].data8[0]];

	hash[2].data16[1] = CONST_BYTE_TO_HEX[hash[1].data8[3]];
	hash[2].data16[0] = CONST_BYTE_TO_HEX[hash[1].data8[2]];
	uint16_t t                      = CONST_BYTE_TO_HEX[hash[1].data8[0]];
	hash[1].data16[1] = CONST_BYTE_TO_HEX[hash[1].data8[1]];
	hash[1].data16[0] = t;
	hash[0].data[0]   = 32;

	doubleMd5Func_Norm(hash, hash);
}

uint32_t DoubleMD5::initCalc()
{
#ifdef ARC_x86
	if ((getInstructionSets() & (IS_AVX | IS_XOP)) == (IS_AVX | IS_XOP))
	{
		return HashFactory::registerHash("doublemd5", QUOTE(__DOUBLE_MD5_INTERLACED__) "xAVX_XOP", 0, 35, DoubleMD5::calcAVXXOP,  16, 4 * __DOUBLE_MD5_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 9, 6);
	}
	else if ((getInstructionSets() & IS_AVX) == IS_AVX)
	{
		return HashFactory::registerHash("doublemd5", QUOTE(__DOUBLE_MD5_INTERLACED__) "xAVX",     0, 35, DoubleMD5::calcAVX,     16, 4 * __DOUBLE_MD5_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 9, 6);
	}
	else if ((getInstructionSets() & IS_SSE2) == IS_SSE2)
	{
		return HashFactory::registerHash("doublemd5", QUOTE(__DOUBLE_MD5_INTERLACED__) "xSSE2",    0, 35, DoubleMD5::calcSSE2,    16, 4 * __DOUBLE_MD5_INTERLACED__, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 9, 6);
	}
#endif
	return HashFactory::registerHash("doublemd5", "scalar",                                    0, 35, DoubleMD5::calcNorm,    16,                             1, CE_ASCII_BIN8 | CE_80_NULL_PADDED, 9, 6);
}
