/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef MD4_HELP_H
#define MD4_HELP_H

#include "common.h"

#ifndef ARC_x86
	// Non-x86... Yeah I got nothing
#else

#ifdef _WIN32
	#include <immintrin.h> // AVX and older
	#include <ammintrin.h> // XOP
#else
	#include <x86intrin.h>
#endif

#ifdef ARC_x86_64
	#define __MD4_INTERLACED__ 4
#elif defined(ARC_x86_32)
	#define __MD4_INTERLACED__ 4
#endif

#if __SHA1_INTERLACED__ > 5
	#undef  __MD4_INTERLACED__
	#define __MD4_INTERLACED__ 5
#endif

#define LOAD_CONST_MD4_A   _mm_set1_epi32(0x67452301)
#define LOAD_CONST_MD4_B   _mm_set1_epi32(0xefcdab89)
#define LOAD_CONST_MD4_C   _mm_set1_epi32(0x98badcfe)
#define LOAD_CONST_MD4_D   _mm_set1_epi32(0x10325476)
#define LOAD_CONST_MD4_GG  _mm_set1_epi32(0x5a827999)
#define LOAD_CONST_MD4_HH  _mm_set1_epi32(0x6ed9eba1)

#define ZERO     _mm_set1_epi32(0)

#define MD4_FF_(  a,b,c,d,x,s,j)      if (__MD4_INTERLACED__ > j) { a[j] =               _mm_add_epi32(_mm_add_epi32(a[j], _mm_or_si128(_mm_and_si128(b[j], c[j]), _mm_andnot_si128(b[j], d[j]))                 ), x(j))                    ; a[j] = ROT(a[j], s); }
#define MD4_GG_(  a,b,c,d,x,s,j)      if (__MD4_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_and_si128(_mm_or_si128(_mm_and_si128(b[j], c[j]), d[j]), _mm_or_si128(b[j], c[j]))), x(j)), LOAD_CONST_MD4_GG); a[j] = ROT(a[j], s); }
#define MD4_HH_(  a,b,c,d,x,s,j)      if (__MD4_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(b[j], _mm_xor_si128(c[j], d[j]))                                        ), x(j)), LOAD_CONST_MD4_HH); a[j] = ROT(a[j], s); }
#define MD4_HH_E_(a,b,c,d,x,s,j,t)    if (__MD4_INTERLACED__ > j) { t[j] = _mm_xor_si128(b[j], c[j]); a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(t[j], d[j])                           ), x(j)), LOAD_CONST_MD4_HH); a[j] = ROT(a[j], s); }
#define MD4_HH_O_(a,b,c,d,x,s,j,t)    if (__MD4_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(a[j], _mm_xor_si128(b[j], t[j])                                                             ), x(j)), LOAD_CONST_MD4_HH); a[j] = ROT(a[j], s); }

#define MD4_FF(  a,b,c,d,x,s)         MD4_FF_(  a,b,c,d,x,s,0);   MD4_FF_(  a,b,c,d,x,s,1);   MD4_FF_(  a,b,c,d,x,s,2);   MD4_FF_(  a,b,c,d,x,s,3);   MD4_FF_(  a,b,c,d,x,s,4);
#define MD4_GG(  a,b,c,d,x,s)         MD4_GG_(  a,b,c,d,x,s,0);   MD4_GG_(  a,b,c,d,x,s,1);   MD4_GG_(  a,b,c,d,x,s,2);   MD4_GG_(  a,b,c,d,x,s,3);   MD4_GG_(  a,b,c,d,x,s,4);
#define MD4_HH(  a,b,c,d,x,s)         MD4_HH_(  a,b,c,d,x,s,0);   MD4_HH_(  a,b,c,d,x,s,1);   MD4_HH_(  a,b,c,d,x,s,2);   MD4_HH_(  a,b,c,d,x,s,3);   MD4_HH_(  a,b,c,d,x,s,4);
#define MD4_HH_E(a,b,c,d,x,s,t)       MD4_HH_E_(a,b,c,d,x,s,0,t); MD4_HH_E_(a,b,c,d,x,s,1,t); MD4_HH_E_(a,b,c,d,x,s,2,t); MD4_HH_E_(a,b,c,d,x,s,3,t); MD4_HH_E_(a,b,c,d,x,s,4,t);
#define MD4_HH_O(a,b,c,d,x,s,t)       MD4_HH_O_(a,b,c,d,x,s,0,t); MD4_HH_O_(a,b,c,d,x,s,1,t); MD4_HH_O_(a,b,c,d,x,s,2,t); MD4_HH_O_(a,b,c,d,x,s,3,t); MD4_HH_O_(a,b,c,d,x,s,4,t);

#define x0(i)   pw[1 * __MD4_INTERLACED__ + i]
#define x1(i)   pw[2 * __MD4_INTERLACED__ + i]
#define x2(i)   pw[3 * __MD4_INTERLACED__ + i]
#define x3(i)   pw[4 * __MD4_INTERLACED__ + i]
#define x4(i)   pw[5 * __MD4_INTERLACED__ + i]
#define x5(i)   pw[6 * __MD4_INTERLACED__ + i]
#define x6(i)   pw[7 * __MD4_INTERLACED__ + i]
#define x7(i)   pw[8 * __MD4_INTERLACED__ + i]
#define x8(i)   ZERO
#define x9(i)   ZERO
#define x10(i)  ZERO
#define x11(i)  ZERO
#define x12(i)  ZERO
#define x13(i)  ZERO
#define x14(i)  pw[0 * __MD4_INTERLACED__ + i]
#define x15(i)  ZERO

#define ROT(n,s) _mm_or_si128(_mm_slli_epi32(n, s), _mm_srli_epi32(n, (32 - s)))
__inline__ void md4Func(__m128i *pw, __m128i *hash)
{
	__m128i a[__MD4_INTERLACED__], b[__MD4_INTERLACED__], c[__MD4_INTERLACED__], d[__MD4_INTERLACED__];

	if (__MD4_INTERLACED__ > 0) { a[0] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 1) { a[1] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 2) { a[2] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 3) { a[3] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 4) { a[4] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 0) { b[0] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 1) { b[1] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 2) { b[2] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 3) { b[3] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 4) { b[4] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 0) { c[0] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 1) { c[1] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 2) { c[2] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 3) { c[3] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 4) { c[4] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 0) { d[0] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 1) { d[1] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 2) { d[2] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 3) { d[3] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 4) { d[4] = LOAD_CONST_MD4_D; }

	if (__MD4_INTERLACED__ > 0) { x14(0) = _mm_slli_epi32(x14(0), 3); }
	if (__MD4_INTERLACED__ > 1) { x14(1) = _mm_slli_epi32(x14(1), 3); }
	if (__MD4_INTERLACED__ > 2) { x14(2) = _mm_slli_epi32(x14(2), 3); }
	if (__MD4_INTERLACED__ > 3) { x14(3) = _mm_slli_epi32(x14(3), 3); }
	if (__MD4_INTERLACED__ > 4) { x14(4) = _mm_slli_epi32(x14(4), 3); }

	MD4_FF(  a,b,c,d, x0, 3    ); MD4_FF(  d,a,b,c, x1, 7    ); MD4_FF(  c,d,a,b, x2,11    ); MD4_FF(  b,c,d,a, x3,19    );
	MD4_FF(  a,b,c,d, x4, 3    ); MD4_FF(  d,a,b,c, x5, 7    ); MD4_FF(  c,d,a,b, x6,11    ); MD4_FF(  b,c,d,a, x7,19    );
	MD4_FF(  a,b,c,d, x8, 3    ); MD4_FF(  d,a,b,c, x9, 7    ); MD4_FF(  c,d,a,b,x10,11    ); MD4_FF(  b,c,d,a,x11,19    );
	MD4_FF(  a,b,c,d,x12, 3    ); MD4_FF(  d,a,b,c,x13, 7    ); MD4_FF(  c,d,a,b,x14,11    ); MD4_FF(  b,c,d,a,x15,19    );

	MD4_GG(  a,b,c,d, x0, 3    ); MD4_GG(  d,a,b,c, x4, 5    ); MD4_GG(  c,d,a,b, x8, 9    ); MD4_GG(  b,c,d,a,x12,13    );
	MD4_GG(  a,b,c,d, x1, 3    ); MD4_GG(  d,a,b,c, x5, 5    ); MD4_GG(  c,d,a,b, x9, 9    ); MD4_GG(  b,c,d,a,x13,13    );
	MD4_GG(  a,b,c,d, x2, 3    ); MD4_GG(  d,a,b,c, x6, 5    ); MD4_GG(  c,d,a,b,x10, 9    ); MD4_GG(  b,c,d,a,x14,13    );
	MD4_GG(  a,b,c,d, x3, 3    ); MD4_GG(  d,a,b,c, x7, 5    ); MD4_GG(  c,d,a,b,x11, 9    ); MD4_GG(  b,c,d,a,x15,13    );

	register __m128i tmp[__MD4_INTERLACED__];
	MD4_HH_E(a,b,c,d, x0, 3,tmp); MD4_HH_O(d,a,b,c, x8, 9,tmp); MD4_HH_E(c,d,a,b, x4,11,tmp); MD4_HH_O(b,c,d,a,x12,15,tmp);
	MD4_HH_E(a,b,c,d, x2, 3,tmp); MD4_HH_O(d,a,b,c,x10, 9,tmp); MD4_HH_E(c,d,a,b, x6,11,tmp); MD4_HH_O(b,c,d,a,x14,15,tmp);
	MD4_HH_E(a,b,c,d, x1, 3,tmp); MD4_HH_O(d,a,b,c, x9, 9,tmp); MD4_HH_E(c,d,a,b, x5,11,tmp); MD4_HH_O(b,c,d,a,x13,15,tmp);
	MD4_HH_E(a,b,c,d, x3, 3,tmp); MD4_HH_O(d,a,b,c,x11, 9,tmp); MD4_HH_E(c,d,a,b, x7,11,tmp); MD4_HH_O(b,c,d,a,x15,15,tmp);

	if (__MD4_INTERLACED__ > 0) { hash[0 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(a[0], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 1) { hash[0 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(a[1], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 2) { hash[0 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(a[2], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 3) { hash[0 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(a[3], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 4) { hash[0 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(a[4], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 0) { hash[1 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(b[0], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 1) { hash[1 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(b[1], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 2) { hash[1 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(b[2], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 3) { hash[1 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(b[3], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 4) { hash[1 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(b[4], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 0) { hash[2 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(c[0], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 1) { hash[2 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(c[1], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 2) { hash[2 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(c[2], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 3) { hash[2 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(c[3], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 4) { hash[2 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(c[4], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 0) { hash[3 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(d[0], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 1) { hash[3 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(d[1], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 2) { hash[3 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(d[2], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 3) { hash[3 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(d[3], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 4) { hash[3 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(d[4], LOAD_CONST_MD4_D); }
}
#undef ROT

#ifdef XOP_CODE
#undef MD4_FF_

#define ROT(n,s)                _mm_roti_epi32(n, s)
#define MD4_FF_(a,b,c,d,x,s,j)  if (__MD4_INTERLACED__ > j) { a[j] = _mm_add_epi32(_mm_add_epi32(a[j], _mm_cmov_si128(c[j], d[j], b[j])), x(j)); a[j] = ROT(a[j], s); }
__inline__ void md4Func_XOP(__m128i *pw, __m128i *hash)
{
	__m128i a[__MD4_INTERLACED__], b[__MD4_INTERLACED__], c[__MD4_INTERLACED__], d[__MD4_INTERLACED__];

	if (__MD4_INTERLACED__ > 0) { a[0] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 1) { a[1] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 2) { a[2] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 3) { a[3] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 4) { a[4] = LOAD_CONST_MD4_A; }
	if (__MD4_INTERLACED__ > 0) { b[0] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 1) { b[1] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 2) { b[2] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 3) { b[3] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 4) { b[4] = LOAD_CONST_MD4_B; }
	if (__MD4_INTERLACED__ > 0) { c[0] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 1) { c[1] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 2) { c[2] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 3) { c[3] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 4) { c[4] = LOAD_CONST_MD4_C; }
	if (__MD4_INTERLACED__ > 0) { d[0] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 1) { d[1] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 2) { d[2] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 3) { d[3] = LOAD_CONST_MD4_D; }
	if (__MD4_INTERLACED__ > 4) { d[4] = LOAD_CONST_MD4_D; }

	if (__MD4_INTERLACED__ > 0) { x14(0) = _mm_slli_epi32(x14(0), 3); }
	if (__MD4_INTERLACED__ > 1) { x14(1) = _mm_slli_epi32(x14(1), 3); }
	if (__MD4_INTERLACED__ > 2) { x14(2) = _mm_slli_epi32(x14(2), 3); }
	if (__MD4_INTERLACED__ > 3) { x14(3) = _mm_slli_epi32(x14(3), 3); }
	if (__MD4_INTERLACED__ > 4) { x14(4) = _mm_slli_epi32(x14(4), 3); }

	MD4_FF(  a,b,c,d, x0, 3    ); MD4_FF(  d,a,b,c, x1, 7    ); MD4_FF(  c,d,a,b, x2,11    ); MD4_FF(  b,c,d,a, x3,19    );
	MD4_FF(  a,b,c,d, x4, 3    ); MD4_FF(  d,a,b,c, x5, 7    ); MD4_FF(  c,d,a,b, x6,11    ); MD4_FF(  b,c,d,a, x7,19    );
	MD4_FF(  a,b,c,d, x8, 3    ); MD4_FF(  d,a,b,c, x9, 7    ); MD4_FF(  c,d,a,b,x10,11    ); MD4_FF(  b,c,d,a,x11,19    );
	MD4_FF(  a,b,c,d,x12, 3    ); MD4_FF(  d,a,b,c,x13, 7    ); MD4_FF(  c,d,a,b,x14,11    ); MD4_FF(  b,c,d,a,x15,19    );

	MD4_GG(  a,b,c,d, x0, 3    ); MD4_GG(  d,a,b,c, x4, 5    ); MD4_GG(  c,d,a,b, x8, 9    ); MD4_GG(  b,c,d,a,x12,13    );
	MD4_GG(  a,b,c,d, x1, 3    ); MD4_GG(  d,a,b,c, x5, 5    ); MD4_GG(  c,d,a,b, x9, 9    ); MD4_GG(  b,c,d,a,x13,13    );
	MD4_GG(  a,b,c,d, x2, 3    ); MD4_GG(  d,a,b,c, x6, 5    ); MD4_GG(  c,d,a,b,x10, 9    ); MD4_GG(  b,c,d,a,x14,13    );
	MD4_GG(  a,b,c,d, x3, 3    ); MD4_GG(  d,a,b,c, x7, 5    ); MD4_GG(  c,d,a,b,x11, 9    ); MD4_GG(  b,c,d,a,x15,13    );

	register __m128i tmp[__MD4_INTERLACED__];
	MD4_HH_E(a,b,c,d, x0, 3,tmp); MD4_HH_O(d,a,b,c, x8, 9,tmp); MD4_HH_E(c,d,a,b, x4,11,tmp); MD4_HH_O(b,c,d,a,x12,15,tmp);
	MD4_HH_E(a,b,c,d, x2, 3,tmp); MD4_HH_O(d,a,b,c,x10, 9,tmp); MD4_HH_E(c,d,a,b, x6,11,tmp); MD4_HH_O(b,c,d,a,x14,15,tmp);
	MD4_HH_E(a,b,c,d, x1, 3,tmp); MD4_HH_O(d,a,b,c, x9, 9,tmp); MD4_HH_E(c,d,a,b, x5,11,tmp); MD4_HH_O(b,c,d,a,x13,15,tmp);
	MD4_HH_E(a,b,c,d, x3, 3,tmp); MD4_HH_O(d,a,b,c,x11, 9,tmp); MD4_HH_E(c,d,a,b, x7,11,tmp); MD4_HH_O(b,c,d,a,x15,15,tmp);

	if (__MD4_INTERLACED__ > 0) { hash[0 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(a[0], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 1) { hash[0 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(a[1], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 2) { hash[0 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(a[2], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 3) { hash[0 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(a[3], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 4) { hash[0 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(a[4], LOAD_CONST_MD4_A); }
	if (__MD4_INTERLACED__ > 0) { hash[1 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(b[0], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 1) { hash[1 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(b[1], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 2) { hash[1 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(b[2], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 3) { hash[1 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(b[3], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 4) { hash[1 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(b[4], LOAD_CONST_MD4_B); }
	if (__MD4_INTERLACED__ > 0) { hash[2 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(c[0], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 1) { hash[2 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(c[1], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 2) { hash[2 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(c[2], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 3) { hash[2 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(c[3], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 4) { hash[2 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(c[4], LOAD_CONST_MD4_C); }
	if (__MD4_INTERLACED__ > 0) { hash[3 * __MD4_INTERLACED__ + 0] = _mm_add_epi32(d[0], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 1) { hash[3 * __MD4_INTERLACED__ + 1] = _mm_add_epi32(d[1], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 2) { hash[3 * __MD4_INTERLACED__ + 2] = _mm_add_epi32(d[2], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 3) { hash[3 * __MD4_INTERLACED__ + 3] = _mm_add_epi32(d[3], LOAD_CONST_MD4_D); }
	if (__MD4_INTERLACED__ > 4) { hash[3 * __MD4_INTERLACED__ + 4] = _mm_add_epi32(d[4], LOAD_CONST_MD4_D); }
}
#undef ROT
#endif

#undef x0
#undef x1
#undef x2
#undef x3
#undef x4
#undef x5
#undef x6
#undef x7
#undef x8
#undef x9
#undef x10
#undef x11
#undef x12
#undef x13
#undef x14
#undef x15

#undef ZERO
#undef MD4_FF
#undef MD4_GG
#undef MD4_HH
#undef MD4_HH_E
#undef MD4_HH_O
#undef MD4_FF_
#undef MD4_GG_
#undef MD4_HH_
#undef MD4_HH_E_
#undef MD4_HH_O_

#undef LOAD_CONST_MD4_A
#undef LOAD_CONST_MD4_B
#undef LOAD_CONST_MD4_C
#undef LOAD_CONST_MD4_D
#undef LOAD_CONST_MD4_GG
#undef LOAD_CONST_MD4_HH

#undef __inline__

#endif
#endif
