/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef SHA1_HELP_H
#define SHA1_HELP_H

#include "common.h"

#ifndef ARC_x86
	// Non-x86... Yeah I got nothing
#else

#ifdef _WIN32
	#include <immintrin.h> // AVX and older
	#include <ammintrin.h> // XOP
#else
	#include <x86intrin.h>
#endif

#ifdef ARC_x86_64
	#define __SHA1_INTERLACED__ 1
#elif defined(ARC_x86_32)
	#define __SHA1_INTERLACED__ 1
#endif

#if __SHA1_INTERLACED__ > 3
	#undef  __SHA1_INTERLACED__
	#define __SHA1_INTERLACED__ 3
#endif

#define LOAD_CONST_SHA1_A   _mm_set1_epi32(0x67452301)
#define LOAD_CONST_SHA1_B   _mm_set1_epi32(0xefcdab89)
#define LOAD_CONST_SHA1_C   _mm_set1_epi32(0x98badcfe)
#define LOAD_CONST_SHA1_D   _mm_set1_epi32(0x10325476)
#define LOAD_CONST_SHA1_E   _mm_set1_epi32(0xc3d2e1f0)
#define LOAD_CONST_SHA1_FF  _mm_set1_epi32(0x5a827999)
#define LOAD_CONST_SHA1_GG  _mm_set1_epi32(0x6ed9eba1)
#define LOAD_CONST_SHA1_HH  _mm_set1_epi32(0x8f1bbcdc)
#define LOAD_CONST_SHA1_II  _mm_set1_epi32(0xca62c1d6)

#define ZERO     _mm_set1_epi32(0)

//#define SHA1_FF_(a,b,c,d,e,x,i,j)    if (__SHA1_INTERLACED__ > j) { e[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(_mm_add_epi32(e[j], _mm_or_si128(_mm_and_si128(b[j], c[j]), _mm_andnot_si128(b[j], d[j]))                 ), x(i,j)), LOAD_CONST_SHA1_FF), ROT(a[j], 5)); b[j] = ROT(b[j], 30); }
#define SHA1_FF_(a,b,c,d,e,x,i,j)      if (__SHA1_INTERLACED__ > j) { e[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(_mm_add_epi32(e[j], _mm_xor_si128(_mm_and_si128(_mm_xor_si128(c[j], d[j]), b[j]), d[j])                   ), x(i,j)), LOAD_CONST_SHA1_FF), ROT(a[j], 5)); b[j] = ROT(b[j], 30); }
#define SHA1_GG_(a,b,c,d,e,x,i,j)      if (__SHA1_INTERLACED__ > j) { e[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(_mm_add_epi32(e[j], _mm_xor_si128(b[j], _mm_xor_si128(c[j], d[j]))                                        ), x(i,j)), LOAD_CONST_SHA1_GG), ROT(a[j], 5)); b[j] = ROT(b[j], 30); }
#define SHA1_HH_(a,b,c,d,e,x,i,j)      if (__SHA1_INTERLACED__ > j) { e[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(_mm_add_epi32(e[j], _mm_or_si128(_mm_and_si128(_mm_or_si128(b[j], c[j]), d[j]), _mm_and_si128(b[j], c[j]))), x(i,j)), LOAD_CONST_SHA1_HH), ROT(a[j], 5)); b[j] = ROT(b[j], 30); }
#define SHA1_II_(a,b,c,d,e,x,i,j)      if (__SHA1_INTERLACED__ > j) { e[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(_mm_add_epi32(e[j], _mm_xor_si128(b[j], _mm_xor_si128(c[j], d[j]))                                        ), x(i,j)), LOAD_CONST_SHA1_II), ROT(a[j], 5)); b[j] = ROT(b[j], 30); }

#define SHA1_FF(a,b,c,d,e,x,i)         SHA1_FF_(  a,b,c,d,e,  x,i,0); SHA1_FF_(  a,b,c,d,e,  x,i,1); SHA1_FF_(  a,b,c,d,e,  x,i,2);
#define SHA1_GG(a,b,c,d,e,x,i)         SHA1_GG_(  a,b,c,d,e,  x,i,0); SHA1_GG_(  a,b,c,d,e,  x,i,1); SHA1_GG_(  a,b,c,d,e,  x,i,2);
#define SHA1_HH(a,b,c,d,e,x,i)         SHA1_HH_(  a,b,c,d,e,  x,i,0); SHA1_HH_(  a,b,c,d,e,  x,i,1); SHA1_HH_(  a,b,c,d,e,  x,i,2);
#define SHA1_II(a,b,c,d,e,x,i)         SHA1_II_(  a,b,c,d,e,  x,i,0); SHA1_II_(  a,b,c,d,e,  x,i,1); SHA1_II_(  a,b,c,d,e,  x,i,2);

#define x0(i,j)    (y(0,j) = SWAP_ENDIAN(pw[1 * __SHA1_INTERLACED__ + j]))
#define x1(i,j)    (y(1,j) = SWAP_ENDIAN(pw[2 * __SHA1_INTERLACED__ + j]))
#define x2(i,j)    (y(2,j) = SWAP_ENDIAN(pw[3 * __SHA1_INTERLACED__ + j]))
#define x3(i,j)    (y(3,j) = SWAP_ENDIAN(pw[4 * __SHA1_INTERLACED__ + j]))
#define x4(i,j)    (y(4,j) = SWAP_ENDIAN(pw[5 * __SHA1_INTERLACED__ + j]))
#define x5(i,j)    (y(5,j) = SWAP_ENDIAN(pw[6 * __SHA1_INTERLACED__ + j]))
#define x6(i,j)    ZERO
#define x7(i,j)    ZERO
#define x8(i,j)    ZERO
#define x9(i,j)    ZERO
#define x10(i,j)   ZERO
#define x11(i,j)   ZERO
#define x12(i,j)   ZERO
#define x13(i,j)   ZERO
#define x14(i,j)   ZERO
#define x15(i,j)   y(15,j)

#define y(i,j)     hash[((i) & 15) * __SHA1_INTERLACED__ + j]

#define x(i,j)     (y( i,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( i,j), y(i+2,j)), y(i+8,j)), y(i+13,j)), 1))
#define x0_2(i,j)  (y( 0,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( 0,j),  y( 2,j)),     ZERO),      ZERO), 1))
#define x1_2(i,j)  (y( 1,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( 1,j),  y( 3,j)),     ZERO),      ZERO), 1))
#define x2_2(i,j)  (y( 2,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( 2,j),  y( 4,j)),     ZERO),   y(15,j)), 1))
#define x3_2(i,j)  (y( 3,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( 3,j),  y( 5,j)),     ZERO),   y( 0,j)), 1))
#define x4_2(i,j)  (y( 4,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( 4,j),     ZERO),     ZERO),   y( 1,j)), 1))
#define x5_2(i,j)  (y( 5,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y( 5,j),     ZERO),     ZERO),   y( 2,j)), 1))
#define x6_2(i,j)  (y( 6,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),     ZERO),   y( 3,j)), 1))
#define x7_2(i,j)  (y( 7,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),  y(15,j)),   y( 4,j)), 1))
#define x8_2(i,j)  (y( 8,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),  y( 0,j)),   y( 5,j)), 1))
#define x9_2(i,j)  (y( 9,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),  y( 1,j)),   y( 6,j)), 1))
#define x10_2(i,j) (y(10,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),  y( 2,j)),   y( 7,j)), 1))
#define x11_2(i,j) (y(11,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),  y( 3,j)),   y( 8,j)), 1))
#define x12_2(i,j) (y(12,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,     ZERO),  y( 4,j)),   y( 9,j)), 1))
#define x13_2(i,j) (y(13,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,  y(15,j)),  y( 5,j)),   y(10,j)), 1))
#define x14_2(i,j) (y(14,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(   ZERO,  y( 0,j)),  y( 6,j)),   y(11,j)), 1))
#define x15_2(i,j) (y(15,j) = ROT(_mm_xor_si128(_mm_xor_si128(_mm_xor_si128(y(15,j),  y( 1,j)),  y( 7,j)),   y(12,j)), 1))

#define ROT(n,s) _mm_or_si128(_mm_slli_epi32(n, s), _mm_srli_epi32(n, (32 - s)))
#define ROT16(n) _mm_shufflelo_epi16(_mm_shufflehi_epi16(n, 0xb1), 0xb1)
#define SWAP_ENDIAN(n) _mm_or_si128(_mm_srli_epi16(ROT16(n), 8), _mm_slli_epi16(ROT16(n), 8))
__inline__ void sha1Func(__m128i *pw, __m128i *hash)
{
	__m128i a[__SHA1_INTERLACED__], b[__SHA1_INTERLACED__], c[__SHA1_INTERLACED__], d[__SHA1_INTERLACED__], e[__SHA1_INTERLACED__];

	if (__SHA1_INTERLACED__ > 0) { a[0] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 1) { a[1] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 2) { a[2] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 0) { b[0] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 1) { b[1] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 2) { b[2] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 0) { c[0] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 1) { c[1] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 2) { c[2] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 0) { d[0] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 1) { d[1] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 2) { d[2] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 0) { e[0] = LOAD_CONST_SHA1_E; }
	if (__SHA1_INTERLACED__ > 1) { e[1] = LOAD_CONST_SHA1_E; }
	if (__SHA1_INTERLACED__ > 2) { e[2] = LOAD_CONST_SHA1_E; }

	if (__SHA1_INTERLACED__ > 0) { x15(15,0) = _mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 0], 3); }
	if (__SHA1_INTERLACED__ > 1) { x15(15,1) = _mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 1], 3); }
	if (__SHA1_INTERLACED__ > 2) { x15(15,2) = _mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 2], 3); }

	SHA1_FF(a,b,c,d,e,   x0, 0); SHA1_FF(e,a,b,c,d,   x1, 1); SHA1_FF(d,e,a,b,c,   x2, 2); SHA1_FF(c,d,e,a,b,   x3, 3); SHA1_FF(b,c,d,e,a,   x4, 4);
	SHA1_FF(a,b,c,d,e,   x5, 5); SHA1_FF(e,a,b,c,d,   x6, 6); SHA1_FF(d,e,a,b,c,   x7, 7); SHA1_FF(c,d,e,a,b,   x8, 8); SHA1_FF(b,c,d,e,a,   x9, 9);
	SHA1_FF(a,b,c,d,e,  x10,10); SHA1_FF(e,a,b,c,d,  x11,11); SHA1_FF(d,e,a,b,c,  x12,12); SHA1_FF(c,d,e,a,b,  x13,13); SHA1_FF(b,c,d,e,a,  x14,14);
	SHA1_FF(a,b,c,d,e,  x15,15); SHA1_FF(e,a,b,c,d, x0_2,16); SHA1_FF(d,e,a,b,c, x1_2,17); SHA1_FF(c,d,e,a,b, x2_2,18); SHA1_FF(b,c,d,e,a, x3_2,19);

	SHA1_GG(a,b,c,d,e, x4_2,20); SHA1_GG(e,a,b,c,d, x5_2,21); SHA1_GG(d,e,a,b,c, x6_2,22); SHA1_GG(c,d,e,a,b, x7_2,23); SHA1_GG(b,c,d,e,a, x8_2,24);
	SHA1_GG(a,b,c,d,e, x9_2,25); SHA1_GG(e,a,b,c,d,x10_2,26); SHA1_GG(d,e,a,b,c,x11_2,27); SHA1_GG(c,d,e,a,b,x12_2,28); SHA1_GG(b,c,d,e,a,x13_2,29);
	SHA1_GG(a,b,c,d,e,x14_2,30); SHA1_GG(e,a,b,c,d,x15_2,31); SHA1_GG(d,e,a,b,c,    x,32); SHA1_GG(c,d,e,a,b,    x,33); SHA1_GG(b,c,d,e,a,    x,34);
	SHA1_GG(a,b,c,d,e,    x,35); SHA1_GG(e,a,b,c,d,    x,36); SHA1_GG(d,e,a,b,c,    x,37); SHA1_GG(c,d,e,a,b,    x,38); SHA1_GG(b,c,d,e,a,    x,39);

	SHA1_HH(a,b,c,d,e,    x,40); SHA1_HH(e,a,b,c,d,    x,41); SHA1_HH(d,e,a,b,c,    x,42); SHA1_HH(c,d,e,a,b,    x,43); SHA1_HH(b,c,d,e,a,    x,44);
	SHA1_HH(a,b,c,d,e,    x,45); SHA1_HH(e,a,b,c,d,    x,46); SHA1_HH(d,e,a,b,c,    x,47); SHA1_HH(c,d,e,a,b,    x,48); SHA1_HH(b,c,d,e,a,    x,49);
	SHA1_HH(a,b,c,d,e,    x,50); SHA1_HH(e,a,b,c,d,    x,51); SHA1_HH(d,e,a,b,c,    x,52); SHA1_HH(c,d,e,a,b,    x,53); SHA1_HH(b,c,d,e,a,    x,54);
	SHA1_HH(a,b,c,d,e,    x,55); SHA1_HH(e,a,b,c,d,    x,56); SHA1_HH(d,e,a,b,c,    x,57); SHA1_HH(c,d,e,a,b,    x,58); SHA1_HH(b,c,d,e,a,    x,59);

	SHA1_II(a,b,c,d,e,    x,60); SHA1_II(e,a,b,c,d,    x,61); SHA1_II(d,e,a,b,c,    x,62); SHA1_II(c,d,e,a,b,    x,63); SHA1_II(b,c,d,e,a,    x,64);
	SHA1_II(a,b,c,d,e,    x,65); SHA1_II(e,a,b,c,d,    x,66); SHA1_II(d,e,a,b,c,    x,67); SHA1_II(c,d,e,a,b,    x,68); SHA1_II(b,c,d,e,a,    x,69);
	SHA1_II(a,b,c,d,e,    x,70); SHA1_II(e,a,b,c,d,    x,71); SHA1_II(d,e,a,b,c,    x,72); SHA1_II(c,d,e,a,b,    x,73); SHA1_II(b,c,d,e,a,    x,74);
	SHA1_II(a,b,c,d,e,    x,75); SHA1_II(e,a,b,c,d,    x,76); SHA1_II(d,e,a,b,c,    x,77); SHA1_II(c,d,e,a,b,    x,78); SHA1_II(b,c,d,e,a,    x,79);

	if (__SHA1_INTERLACED__ > 0) { y(0,0) = SWAP_ENDIAN(_mm_add_epi32(a[0], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 1) { y(0,1) = SWAP_ENDIAN(_mm_add_epi32(a[1], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 2) { y(0,2) = SWAP_ENDIAN(_mm_add_epi32(a[2], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 0) { y(1,0) = SWAP_ENDIAN(_mm_add_epi32(b[0], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 1) { y(1,1) = SWAP_ENDIAN(_mm_add_epi32(b[1], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 2) { y(1,2) = SWAP_ENDIAN(_mm_add_epi32(b[2], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 0) { y(2,0) = SWAP_ENDIAN(_mm_add_epi32(c[0], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 1) { y(2,1) = SWAP_ENDIAN(_mm_add_epi32(c[1], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 2) { y(2,2) = SWAP_ENDIAN(_mm_add_epi32(c[2], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 0) { y(3,0) = SWAP_ENDIAN(_mm_add_epi32(d[0], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 1) { y(3,1) = SWAP_ENDIAN(_mm_add_epi32(d[1], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 2) { y(3,2) = SWAP_ENDIAN(_mm_add_epi32(d[2], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 0) { y(4,0) = SWAP_ENDIAN(_mm_add_epi32(e[0], LOAD_CONST_SHA1_E)); }
	if (__SHA1_INTERLACED__ > 1) { y(4,1) = SWAP_ENDIAN(_mm_add_epi32(e[1], LOAD_CONST_SHA1_E)); }
	if (__SHA1_INTERLACED__ > 2) { y(4,2) = SWAP_ENDIAN(_mm_add_epi32(e[2], LOAD_CONST_SHA1_E)); }
}
#undef ROT
#undef ROT16
#undef SWAP_ENDIAN

#define ROT(n,s)       _mm_or_si128(_mm_slli_epi32(n, s), _mm_srli_epi32(n, (32 - s)))
#define SWAP_ENDIAN(n) _mm_shuffle_epi8(n, _mm_set_epi32(0x0c0d0e0f,0x08090a0b,0x04050607,0x00010203))
__inline__ void sha1Func_SSSE3(__m128i *pw, __m128i *hash)
{
	__m128i a[__SHA1_INTERLACED__], b[__SHA1_INTERLACED__], c[__SHA1_INTERLACED__], d[__SHA1_INTERLACED__], e[__SHA1_INTERLACED__];

	if (__SHA1_INTERLACED__ > 0) { a[0] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 1) { a[1] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 2) { a[2] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 0) { b[0] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 1) { b[1] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 2) { b[2] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 0) { c[0] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 1) { c[1] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 2) { c[2] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 0) { d[0] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 1) { d[1] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 2) { d[2] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 0) { e[0] = LOAD_CONST_SHA1_E; }
	if (__SHA1_INTERLACED__ > 1) { e[1] = LOAD_CONST_SHA1_E; }
	if (__SHA1_INTERLACED__ > 2) { e[2] = LOAD_CONST_SHA1_E; }

	if (__SHA1_INTERLACED__ > 0) { x15(15,0) = _mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 0], 3); }
	if (__SHA1_INTERLACED__ > 1) { x15(15,1) = _mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 1], 3); }
	if (__SHA1_INTERLACED__ > 2) { x15(15,2) = _mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 2], 3); }

	SHA1_FF(a,b,c,d,e,   x0, 0); SHA1_FF(e,a,b,c,d,   x1, 1); SHA1_FF(d,e,a,b,c,   x2, 2); SHA1_FF(c,d,e,a,b,   x3, 3); SHA1_FF(b,c,d,e,a,   x4, 4);
	SHA1_FF(a,b,c,d,e,   x5, 5); SHA1_FF(e,a,b,c,d,   x6, 6); SHA1_FF(d,e,a,b,c,   x7, 7); SHA1_FF(c,d,e,a,b,   x8, 8); SHA1_FF(b,c,d,e,a,   x9, 9);
	SHA1_FF(a,b,c,d,e,  x10,10); SHA1_FF(e,a,b,c,d,  x11,11); SHA1_FF(d,e,a,b,c,  x12,12); SHA1_FF(c,d,e,a,b,  x13,13); SHA1_FF(b,c,d,e,a,  x14,14);
	SHA1_FF(a,b,c,d,e,  x15,15); SHA1_FF(e,a,b,c,d, x0_2,16); SHA1_FF(d,e,a,b,c, x1_2,17); SHA1_FF(c,d,e,a,b, x2_2,18); SHA1_FF(b,c,d,e,a, x3_2,19);

	SHA1_GG(a,b,c,d,e, x4_2,20); SHA1_GG(e,a,b,c,d, x5_2,21); SHA1_GG(d,e,a,b,c, x6_2,22); SHA1_GG(c,d,e,a,b, x7_2,23); SHA1_GG(b,c,d,e,a, x8_2,24);
	SHA1_GG(a,b,c,d,e, x9_2,25); SHA1_GG(e,a,b,c,d,x10_2,26); SHA1_GG(d,e,a,b,c,x11_2,27); SHA1_GG(c,d,e,a,b,x12_2,28); SHA1_GG(b,c,d,e,a,x13_2,29);
	SHA1_GG(a,b,c,d,e,x14_2,30); SHA1_GG(e,a,b,c,d,x15_2,31); SHA1_GG(d,e,a,b,c,    x,32); SHA1_GG(c,d,e,a,b,    x,33); SHA1_GG(b,c,d,e,a,    x,34);
	SHA1_GG(a,b,c,d,e,    x,35); SHA1_GG(e,a,b,c,d,    x,36); SHA1_GG(d,e,a,b,c,    x,37); SHA1_GG(c,d,e,a,b,    x,38); SHA1_GG(b,c,d,e,a,    x,39);

	SHA1_HH(a,b,c,d,e,    x,40); SHA1_HH(e,a,b,c,d,    x,41); SHA1_HH(d,e,a,b,c,    x,42); SHA1_HH(c,d,e,a,b,    x,43); SHA1_HH(b,c,d,e,a,    x,44);
	SHA1_HH(a,b,c,d,e,    x,45); SHA1_HH(e,a,b,c,d,    x,46); SHA1_HH(d,e,a,b,c,    x,47); SHA1_HH(c,d,e,a,b,    x,48); SHA1_HH(b,c,d,e,a,    x,49);
	SHA1_HH(a,b,c,d,e,    x,50); SHA1_HH(e,a,b,c,d,    x,51); SHA1_HH(d,e,a,b,c,    x,52); SHA1_HH(c,d,e,a,b,    x,53); SHA1_HH(b,c,d,e,a,    x,54);
	SHA1_HH(a,b,c,d,e,    x,55); SHA1_HH(e,a,b,c,d,    x,56); SHA1_HH(d,e,a,b,c,    x,57); SHA1_HH(c,d,e,a,b,    x,58); SHA1_HH(b,c,d,e,a,    x,59);

	SHA1_II(a,b,c,d,e,    x,60); SHA1_II(e,a,b,c,d,    x,61); SHA1_II(d,e,a,b,c,    x,62); SHA1_II(c,d,e,a,b,    x,63); SHA1_II(b,c,d,e,a,    x,64);
	SHA1_II(a,b,c,d,e,    x,65); SHA1_II(e,a,b,c,d,    x,66); SHA1_II(d,e,a,b,c,    x,67); SHA1_II(c,d,e,a,b,    x,68); SHA1_II(b,c,d,e,a,    x,69);
	SHA1_II(a,b,c,d,e,    x,70); SHA1_II(e,a,b,c,d,    x,71); SHA1_II(d,e,a,b,c,    x,72); SHA1_II(c,d,e,a,b,    x,73); SHA1_II(b,c,d,e,a,    x,74);
	SHA1_II(a,b,c,d,e,    x,75); SHA1_II(e,a,b,c,d,    x,76); SHA1_II(d,e,a,b,c,    x,77); SHA1_II(c,d,e,a,b,    x,78); SHA1_II(b,c,d,e,a,    x,79);

	if (__SHA1_INTERLACED__ > 0) { y(0,0) = SWAP_ENDIAN(_mm_add_epi32(a[0], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 1) { y(0,1) = SWAP_ENDIAN(_mm_add_epi32(a[1], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 2) { y(0,2) = SWAP_ENDIAN(_mm_add_epi32(a[2], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 0) { y(1,0) = SWAP_ENDIAN(_mm_add_epi32(b[0], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 1) { y(1,1) = SWAP_ENDIAN(_mm_add_epi32(b[1], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 2) { y(1,2) = SWAP_ENDIAN(_mm_add_epi32(b[2], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 0) { y(2,0) = SWAP_ENDIAN(_mm_add_epi32(c[0], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 1) { y(2,1) = SWAP_ENDIAN(_mm_add_epi32(c[1], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 2) { y(2,2) = SWAP_ENDIAN(_mm_add_epi32(c[2], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 0) { y(3,0) = SWAP_ENDIAN(_mm_add_epi32(d[0], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 1) { y(3,1) = SWAP_ENDIAN(_mm_add_epi32(d[1], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 2) { y(3,2) = SWAP_ENDIAN(_mm_add_epi32(d[2], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 0) { y(4,0) = SWAP_ENDIAN(_mm_add_epi32(e[0], LOAD_CONST_SHA1_E)); }
	if (__SHA1_INTERLACED__ > 1) { y(4,1) = SWAP_ENDIAN(_mm_add_epi32(e[1], LOAD_CONST_SHA1_E)); }
	if (__SHA1_INTERLACED__ > 2) { y(4,2) = SWAP_ENDIAN(_mm_add_epi32(e[2], LOAD_CONST_SHA1_E)); }
}
#undef ROT
#undef SWAP_ENDIAN

#ifdef XOP_CODE
#undef SHA1_FF_

#define ROT(n,s)                   _mm_roti_epi32(n, s)
//#define SWAP_ENDIAN(n)             _mm_shuffle_epi8(n, CONST_SHA1_SWAP_ENDIAN)
#define SWAP_ENDIAN(n)             _mm_or_si128(_mm_srli_epi16(_mm_roti_epi32(n, 16), 8), _mm_slli_epi16(_mm_roti_epi32(n, 16), 8))
#define SHA1_FF_(a,b,c,d,e,x,i,j)  if (__SHA1_INTERLACED__ > j) { e[j] = _mm_add_epi32(_mm_add_epi32(_mm_add_epi32(_mm_add_epi32(e[j], _mm_cmov_si128(c[j], d[j], b[j])), x(i,j)), LOAD_CONST_SHA1_FF), ROT(a[j], 5)); b[j] = ROT(b[j], 30); }

__inline__ void sha1Func_XOP(__m128i *pw, __m128i *hash)
{
	__m128i a[__SHA1_INTERLACED__], b[__SHA1_INTERLACED__], c[__SHA1_INTERLACED__], d[__SHA1_INTERLACED__], e[__SHA1_INTERLACED__];

	if (__SHA1_INTERLACED__ > 0) { a[0] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 1) { a[1] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 2) { a[2] = LOAD_CONST_SHA1_A; }
	if (__SHA1_INTERLACED__ > 0) { b[0] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 1) { b[1] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 2) { b[2] = LOAD_CONST_SHA1_B; }
	if (__SHA1_INTERLACED__ > 0) { c[0] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 1) { c[1] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 2) { c[2] = LOAD_CONST_SHA1_C; }
	if (__SHA1_INTERLACED__ > 0) { d[0] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 1) { d[1] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 2) { d[2] = LOAD_CONST_SHA1_D; }
	if (__SHA1_INTERLACED__ > 0) { e[0] = LOAD_CONST_SHA1_E; }
	if (__SHA1_INTERLACED__ > 1) { e[1] = LOAD_CONST_SHA1_E; }
	if (__SHA1_INTERLACED__ > 2) { e[2] = LOAD_CONST_SHA1_E; }

	if (__SHA1_INTERLACED__ > 0) { x15(0,0) = SWAP_ENDIAN(_mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 0], 3)); }
	if (__SHA1_INTERLACED__ > 1) { x15(0,1) = SWAP_ENDIAN(_mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 1], 3)); }
	if (__SHA1_INTERLACED__ > 2) { x15(0,2) = SWAP_ENDIAN(_mm_slli_epi32(pw[0 * __SHA1_INTERLACED__ + 2], 3)); }

	SHA1_FF(a,b,c,d,e,   x0, 0); SHA1_FF(e,a,b,c,d,   x1, 1); SHA1_FF(d,e,a,b,c,   x2, 2); SHA1_FF(c,d,e,a,b,   x3, 3); SHA1_FF(b,c,d,e,a,   x4, 4);
	SHA1_FF(a,b,c,d,e,   x5, 5); SHA1_FF(e,a,b,c,d,   x6, 6); SHA1_FF(d,e,a,b,c,   x7, 7); SHA1_FF(c,d,e,a,b,   x8, 8); SHA1_FF(b,c,d,e,a,   x9, 9);
	SHA1_FF(a,b,c,d,e,  x10,10); SHA1_FF(e,a,b,c,d,  x11,11); SHA1_FF(d,e,a,b,c,  x12,12); SHA1_FF(c,d,e,a,b,  x13,13); SHA1_FF(b,c,d,e,a,  x14,14);
	SHA1_FF(a,b,c,d,e,  x15,15); SHA1_FF(e,a,b,c,d, x0_2,16); SHA1_FF(d,e,a,b,c, x1_2,17); SHA1_FF(c,d,e,a,b, x2_2,18); SHA1_FF(b,c,d,e,a, x3_2,19);

	SHA1_GG(a,b,c,d,e, x4_2,20); SHA1_GG(e,a,b,c,d, x5_2,21); SHA1_GG(d,e,a,b,c, x6_2,22); SHA1_GG(c,d,e,a,b, x7_2,23); SHA1_GG(b,c,d,e,a, x8_2,24);
	SHA1_GG(a,b,c,d,e, x9_2,25); SHA1_GG(e,a,b,c,d,x10_2,26); SHA1_GG(d,e,a,b,c,x11_2,27); SHA1_GG(c,d,e,a,b,x12_2,28); SHA1_GG(b,c,d,e,a,x13_2,29);
	SHA1_GG(a,b,c,d,e,x14_2,30); SHA1_GG(e,a,b,c,d,x15_2,31); SHA1_GG(d,e,a,b,c,    x,32); SHA1_GG(c,d,e,a,b,    x,33); SHA1_GG(b,c,d,e,a,    x,34);
	SHA1_GG(a,b,c,d,e,    x,35); SHA1_GG(e,a,b,c,d,    x,36); SHA1_GG(d,e,a,b,c,    x,37); SHA1_GG(c,d,e,a,b,    x,38); SHA1_GG(b,c,d,e,a,    x,39);

	SHA1_HH(a,b,c,d,e,    x,40); SHA1_HH(e,a,b,c,d,    x,41); SHA1_HH(d,e,a,b,c,    x,42); SHA1_HH(c,d,e,a,b,    x,43); SHA1_HH(b,c,d,e,a,    x,44);
	SHA1_HH(a,b,c,d,e,    x,45); SHA1_HH(e,a,b,c,d,    x,46); SHA1_HH(d,e,a,b,c,    x,47); SHA1_HH(c,d,e,a,b,    x,48); SHA1_HH(b,c,d,e,a,    x,49);
	SHA1_HH(a,b,c,d,e,    x,50); SHA1_HH(e,a,b,c,d,    x,51); SHA1_HH(d,e,a,b,c,    x,52); SHA1_HH(c,d,e,a,b,    x,53); SHA1_HH(b,c,d,e,a,    x,54);
	SHA1_HH(a,b,c,d,e,    x,55); SHA1_HH(e,a,b,c,d,    x,56); SHA1_HH(d,e,a,b,c,    x,57); SHA1_HH(c,d,e,a,b,    x,58); SHA1_HH(b,c,d,e,a,    x,59);

	SHA1_II(a,b,c,d,e,    x,60); SHA1_II(e,a,b,c,d,    x,61); SHA1_II(d,e,a,b,c,    x,62); SHA1_II(c,d,e,a,b,    x,63); SHA1_II(b,c,d,e,a,    x,64);
	SHA1_II(a,b,c,d,e,    x,65); SHA1_II(e,a,b,c,d,    x,66); SHA1_II(d,e,a,b,c,    x,67); SHA1_II(c,d,e,a,b,    x,68); SHA1_II(b,c,d,e,a,    x,69);
	SHA1_II(a,b,c,d,e,    x,70); SHA1_II(e,a,b,c,d,    x,71); SHA1_II(d,e,a,b,c,    x,72); SHA1_II(c,d,e,a,b,    x,73); SHA1_II(b,c,d,e,a,    x,74);
	SHA1_II(a,b,c,d,e,    x,75); SHA1_II(e,a,b,c,d,    x,76); SHA1_II(d,e,a,b,c,    x,77); SHA1_II(c,d,e,a,b,    x,78); SHA1_II(b,c,d,e,a,    x,79);

	if (__SHA1_INTERLACED__ > 0) { y(0,0) = SWAP_ENDIAN(_mm_add_epi32(a[0], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 1) { y(0,1) = SWAP_ENDIAN(_mm_add_epi32(a[1], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 2) { y(0,2) = SWAP_ENDIAN(_mm_add_epi32(a[2], LOAD_CONST_SHA1_A)); }
	if (__SHA1_INTERLACED__ > 0) { y(1,0) = SWAP_ENDIAN(_mm_add_epi32(b[0], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 1) { y(1,1) = SWAP_ENDIAN(_mm_add_epi32(b[1], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 2) { y(1,2) = SWAP_ENDIAN(_mm_add_epi32(b[2], LOAD_CONST_SHA1_B)); }
	if (__SHA1_INTERLACED__ > 0) { y(2,0) = SWAP_ENDIAN(_mm_add_epi32(c[0], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 1) { y(2,1) = SWAP_ENDIAN(_mm_add_epi32(c[1], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 2) { y(2,2) = SWAP_ENDIAN(_mm_add_epi32(c[2], LOAD_CONST_SHA1_C)); }
	if (__SHA1_INTERLACED__ > 0) { y(3,0) = SWAP_ENDIAN(_mm_add_epi32(d[0], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 1) { y(3,1) = SWAP_ENDIAN(_mm_add_epi32(d[1], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 2) { y(3,2) = SWAP_ENDIAN(_mm_add_epi32(d[2], LOAD_CONST_SHA1_D)); }
	if (__SHA1_INTERLACED__ > 0) { y(4,0) = SWAP_ENDIAN(_mm_add_epi32(e[0], LOAD_CONST_SHA1_E)); }
	if (__SHA1_INTERLACED__ > 1) { y(4,1) = SWAP_ENDIAN(_mm_add_epi32(e[1], LOAD_CONST_SHA1_E)); }
	if (__SHA1_INTERLACED__ > 2) { y(4,2) = SWAP_ENDIAN(_mm_add_epi32(e[2], LOAD_CONST_SHA1_E)); }
}
#undef ROT
#undef SWAP_ENDIAN
#endif

#undef x0
#undef x1
#undef x2
#undef x3
#undef x4
#undef x5
#undef x6
#undef x7
#undef x8
#undef x9
#undef x10
#undef x11
#undef x12
#undef x13
#undef x14
#undef x15

#undef x
#undef y

#undef x0_2
#undef x1_2
#undef x2_2
#undef x3_2
#undef x4_2
#undef x5_2
#undef x6_2
#undef x7_2
#undef x8_2
#undef x9_2
#undef x10_2
#undef x11_2
#undef x12_2
#undef x13_2
#undef x14_2
#undef x15_2

#undef ZERO
#undef SHA1_FF
#undef SHA1_GG
#undef SHA1_HH
#undef SHA1_II
#undef SHA1_FF_
#undef SHA1_GG_
#undef SHA1_HH_
#undef SHA1_II_

#undef LOAD_CONST_SHA1_A
#undef LOAD_CONST_SHA1_B
#undef LOAD_CONST_SHA1_C
#undef LOAD_CONST_SHA1_D
#undef LOAD_CONST_SHA1_E
#undef LOAD_CONST_SHA1_FF
#undef LOAD_CONST_SHA1_GG
#undef LOAD_CONST_SHA1_HH
#undef LOAD_CONST_SHA1_II

#undef __inline__

#endif
#endif
