/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef KEY_SPACE_H
#define KEY_SPACE_H

#include "common.h"
#include <list>
#include <string>

using namespace std;

class KeySpace
{
public:
	KeySpace(const KeySpace &ks);
	KeySpace(const char *fileName, uint32_t charaterEncoding);
	KeySpace(list<list<string> > subKeySpaces, uint32_t charaterEncoding);
	virtual ~KeySpace();

	KeySpace& operator=(const KeySpace &ks);

	bool reset(uint64_t pwNum);
	uint32_t next(vector4 *pwOut, uint32_t numPws, uint32_t vecSize);
	uint32_t get(vector4 *pwOut, uint32_t numPws, uint32_t vecSize, uint64_t *pwNums);
	const uint64_t getKeySpace();
	const uint32_t getMaxPwLen();
	const uint32_t getMinPwLen();

protected:
	void init(list<list<string> > subKeySpaces, uint32_t charaterEncoding);
	void cleanUp();
	void copyObject(const KeySpace &ks);
	inline const bool reset(uint64_t pwNum, uint32_t *pwOut, uint32_t *charSetPoses, uint32_t &curSubKeySpace);
	bool cycleLastChar();

	uint64_t m_keySpace;
	uint64_t *m_keySpaceBoundaries;
	uint32_t *m_passwordLengths, *m_divShortCut, *m_charSetPoses, **m_charSetLengths, *m_pw, *m_pwTemp, *m_charSetPosesTemp;
	int *m_numShortCutChars;
	char *m_charSet, ***m_subKeySpaces;
	uint32_t m_numSubKeySpaces, m_curSubKeySpace, m_maxPwLen, m_minPwLen;
	uint32_t m_charaterEncoding, m_offset, m_pwBlocks;
};

#endif
